// Debug version of malloc/free (catching buffer overflows)
// (c) Peter Kankowski, 2008. http://smallcode.weblogs.us/ mailto:kankowski@narod.ru

#include <windows.h>
#include <assert.h>

VOID* DbgMalloc(SIZE_T size) {
	SYSTEM_INFO sys_info;
	GetSystemInfo(&sys_info);

	INT_PTR page = sys_info.dwPageSize;
	assert((page & (page - 1)) == 0); // page size is the power of two
	SIZE_T rounded_size = (size + page - 1) & (-page); // round up to the page boundary

	BYTE* start = (BYTE*)VirtualAlloc(NULL, rounded_size + page, MEM_COMMIT, PAGE_READWRITE);
	DWORD old_protect;
	BOOL res = VirtualProtect(start + rounded_size, page, PAGE_NOACCESS, &old_protect);
	assert(res); UNREFERENCED_PARAMETER(res);
	return start + (rounded_size - size);
}

VOID DbgFree(VOID* memblock) {
	VirtualFree(memblock, 0, MEM_RELEASE);
}

VOID TestDbgMalloc(SIZE_T size) {
	assert(size >= 1);
	static BYTE read;
	BYTE* mem = (BYTE*)DbgMalloc(size);
	*mem = 'x';
	read = *mem;

	*(mem+size-1) = 'y';
	read = *(mem+size-1);

	BOOL caught = FALSE;
	__try {
		read = *(mem+size);
	} __except(EXCEPTION_EXECUTE_HANDLER) {
		caught = TRUE;
	}
	assert(caught);

	caught = FALSE;
	__try {
		*(mem+size) = 'z';
	} __except(EXCEPTION_EXECUTE_HANDLER) {
		caught = TRUE;
	}
	assert(caught);
}

int main(void) {
	TestDbgMalloc(200);
	TestDbgMalloc(9000);
	return 0;
}
